//------------------------------------------------------------------------------
//
//  npgl.c
//
// ANTZ - 3D data viewer with realtime real-world interaction based on NPE.
//
// ANTZ is hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
// Written in 2010-2011 by Shane Saxon - contact@saxondigital.net
//
// Additional code contributed by: Mark Buchanan and Lucas Erickson.
//
// To the extent possible under law, the author(s) have dedicated all copyright 
// and related and neighboring rights to this software to the public domain
// worldwide. This software is distributed without any warranty.
//
// Released under the CC0 license, which is GPL compatible.
//
// You should have received a copy of the CC0 Public Domain Dedication along
// with this software (license file named COPYING.txt). If not, see
// http://creativecommons.org/publicdomain/zero/1.0/
//
//------------------------------------------------------------------------------

#include "npgl.h"

#include "gl/soil/src/SOIL.h"

#include "../npctrl.h"
#include "../os/npos.h"

#include "watchtools.h"


void npGLResizeScene (int width, int height);
void npCopyGPUtoCPU (void* dataRef);


void npDrawTextLabels();										//zz-CAMERA


//------------------------------------------------------------------------------
void npLoadTextures(void* dataRef) 
{
#define	kNPtextureMax 256

	int fileNumber = 1;
	char* filename = (char*)malloc(2048);

	unsigned int textureID;									// texture, debug zz


	pData data = (pData) dataRef;

	printf ("\nSearching For Textures\n");
	// load our texture		// texture, debug zz
	for (fileNumber = 1; fileNumber <= kNPtextureMax; fileNumber++)
	{
		//append the file number, deals with the leading zeros
		if (fileNumber < 10)
			sprintf (filename, "antzmaps/map0000%d.jpg", fileNumber);
		else if (fileNumber < 100)
			sprintf (filename, "antzmaps/map000%d.jpg", fileNumber);
		else if (fileNumber < 1000)
			sprintf (filename, "antzmaps/map00%d.jpg", fileNumber);
		else if (fileNumber < 10000)
			sprintf (filename, "antzmaps/map0%d.jpg", fileNumber);

		textureID = SOIL_load_OGL_texture
		(
			filename,
			SOIL_LOAD_AUTO,
			SOIL_CREATE_NEW_ID,
			SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y
			| SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
		);																// texture, debug zz
		
		//the last texture loaded is the texture count, non-loads return a texture=0
		if (textureID)
		{
			data->io.gl.textureCount = textureID;
			printf ("Loaded textureID: %d\n", textureID);
		}
	}

	if (data->io.gl.textureCount)
		printf ("Done Loading Textures\n\n");
	else
		printf ("Could Not Find Textures\n\n");

}


//------------------------------------------------------------------------------
void npInitGL(void* dataRef) 
{
	npInitGLDraw (dataRef);

	npInitGLPrimitive (dataRef);

	npLoadTextures (dataRef);

//	glShadeModel (GL_FLAT);
//	glDepthFunc (GL_LEQUAL);
//	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
}

//------------------------------------------------------------------------------
void npCloseGL (void* dataRef) 
{
//	npFreeTextures (dataRef);		//debug, zz

	return;
}


// currently just single window support, add multi-screen capability later... debug zz
// redbook has info on how to calculate the projections
//------------------------------------------------------------------------------
void npGLResizeScene (int width, int height)							
{
	int camIndex = 1;	//update with list in data->io->view.camIndex[]		debug zz
	GLfloat ratio = 1.0;

	pData data = npGetDataRef();

	NPnodePtr camNode = data->map.node[camIndex];
	NPcameraPtr camData = camNode->data;

	data->io.gl.width = width;
	data->io.gl.height = height;

	if (height > 0)							//prevent divide by zero
		camData->aspectRatio = (GLfloat)width / (GLfloat)height;

	// Reset the coordinate system before modifying
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	
	// Set the viewport to be the entire window
    glViewport(0, 0, width, height);

	// Set the clipping volume
	gluPerspective (camData->fov, camData->aspectRatio, 
					camData->clipNear, camData->clipFar);
	glMatrixMode (GL_MODELVIEW);

	npGLDrawScene (data);
}

//setup lighting, alpha mode transparency and background
//------------------------------------------------------------------------------
void npGLLighting (void* dataRef)
{
	GLfloat positionA[] = { -1500.0f, 1200.0f, -1800.0f, 1.0f };
	GLfloat positionB[] = { 1500.0f, -1200.0f, 1800.0f, 1.0f };

	GLfloat white[] = {1.0f, 1.0f, 1.0f};
	GLfloat black[] = {0.0f, 0.0f, 0.0f};

	//see old versions, early 2011, for interesting materials lighting
//	GLfloat mShininess[] = {128}; //set the shininess of the material
//	GLfloat globalAmbient[] = { 0.1f, 0.1f, 0.1f, 1.0f };
//	GLfloat redNotGreenNotBlue[] = {1.0f, -1.0f, -1.0f}; //negative light, red

	pData data = (pData) dataRef;

	//needed for transparency
	glEnable (GL_DEPTH_TEST);		//MB-Transp
	glDepthFunc (GL_LESS);

	if( data->io.gl.pickPass )
	{
		glDisable (GL_LIGHTING);		//disable lighting and alpha for pickPass
		glDisable (GL_BLEND);
		glClearColor(0.0f, 0.0f, 0.0f, 1.0f); //background id 0 mapped to black
	}
	else
	{
		glEnable (GL_LIGHTING);		//MB-Transp
		glEnable (GL_BLEND);
		glClearColor(data->io.clear.r, data->io.clear.g, data->io.clear.b, 1.0f);
	}

	//clear background using the clear color, set to black for pickPass
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//set the alphamode
	switch (data->io.gl.alphaMode)
	{
		case kNPalphaModeNull :
			glDisable (GL_BLEND);		//no transparency, disable GL_BLEND
			break;
		case kNPalphaModeSubtractive :
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	//MB-Transp
			break;
		case kNPalphaModeAdditive :
			glBlendFunc (GL_SRC_ALPHA, GL_DST_COLOR);
			break;
		case kNPalphaModeDark :
			glBlendFunc (GL_SRC_COLOR, GL_ONE_MINUS_SRC_ALPHA);
			break;
		default :
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	//MB-Transp
			break;
	}

	glEnable (GL_COLOR_MATERIAL);	//without everything is grayscale

	glEnable (GL_LIGHT0);
	glEnable (GL_LIGHT1);

	glLightfv (GL_LIGHT0, GL_POSITION, positionA);
	glLightfv (GL_LIGHT0, GL_SPECULAR, white);
    glLightfv (GL_LIGHT0, GL_AMBIENT, black);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, white);

	glLightfv (GL_LIGHT1, GL_POSITION, positionB);
	glLightfv (GL_LIGHT1, GL_SPECULAR, white);
    glLightfv (GL_LIGHT1, GL_AMBIENT, black);
    glLightfv (GL_LIGHT1, GL_DIFFUSE, white);
}


//------------------------------------------------------------------------------
void npGLShading (dataRef)
{
	pData data = (pData) dataRef;

	
	//do not render back-faces, faster
	glEnable (GL_CULL_FACE);

	//set shading for flat or smooth
	if (data->io.gl.shade == kNPglShadeSmooth)
		glShadeModel (GL_SMOOTH);
	else
		glShadeModel (GL_FLAT);

	//recalculate normals for non-uniform scaling or rescale normals for speed
	//if disabled then small objects appear brighter and large dimmer
	switch (data->io.gl.normal)
	{
		case kNPglNormalize :				//geo-computationally slow
	//		glDisable (RESCALE_NORMAL);
			glEnable (GL_NORMALIZE);	//recalc normals for non-uniform scaling
			break;
		case kNPglRescaleNormal :			//fast								//debug, zz
			glDisable (GL_NORMALIZE);
			data->io.gl.normal++;			//workaround until RESCALE_NORMAL fixed, zz
	//		glEnable (RESCALE_NORMAL);	//rescale normals, assumes uniform scaling
			break;
		default :							//fastest
			glDisable (GL_NORMALIZE);
	//		glDisable (RESCALE_NORMAL);
			break;
	}
}

//------------------------------------------------------------------------------
void npGLDrawScene (void* dataRef)
{
	//MB-LABEL
	NPfloatXYZ	trans;
	NPfloatXYZA rot;
	//MB-END
	
	int	camIndex = 1;
	GLfloat upX = 0.0f, upY = 0.0f, upZ = 1.0f;
	GLfloat angle = 0.0;

	pData data = (pData) dataRef;
	NPnodePtr camNode = cameraSelected();

	
	//update data, positions, physics, interactions...
	npUpdateCtrl (dataRef);

	npGLLighting (dataRef);

	npGLShading (dataRef);

	npDrawNodes (dataRef);				//draws mesh, graphs, pins, etc...
	
	//npDrawHUD (dataRef);

	//npGLCamera (dataRef);

	glLoadIdentity();
	//MB-LABEL
	
	trans.x = 0.0f;	
	trans.y = 0.0f;
	trans.z = 100.0f;
	rot.x	= 0.0f;
	rot.y	= 0.0f;
	rot.z	= -1.0f;
	rot.angle = 0.0f;
	
#define BUGFIX_MATRIX 0.0000000001
	//HUD LookAt Camera
	glPushMatrix();
	gluLookAt (trans.x, trans.y, trans.z, 
			   trans.x + rot.x,
			   trans.x + rot.y + BUGFIX_MATRIX,
			   trans.z + rot.z,
			   upX, upY, upZ );
//	npDrawTextLabels();				// draws overlays, grids, text... //MB-LABEL

//	glOrtho ( -2.0f, 2.0f, -2.0f, 2.0f, 0.05f, 2.0f);
//	gluOrtho2D ( -2.0f, 2.0f, -2.0f, 2.0f);	
//	npDrawHUD (dataRef);				// draws overlays, grids, text...

/*																			//debug zz
	setOrthographicProjection();
	renderBitmapString(5,30,GLUT_BITMAP_HELVETICA_18,"Lighthouse3D");
	restorePerspectiveProjection();
*/	
	glPopMatrix();
	//MB-LABEL-END	
	
	//3D Camera	
	gluLookAt (	camNode->translate.x, camNode->translate.y, camNode->translate.z,
				camNode->translate.x + camNode->rotateVec.x,
				camNode->translate.y + camNode->rotateVec.y + BUGFIX_MATRIX,
				camNode->translate.z + camNode->rotateVec.z,
				upX, upY, upZ );

	// copy display buffer to main CPU memory
	npCopyGPUtoCPU (dataRef);
}


// copy the rasterized display to the main system memory (CPU RAM or Drive)
//------------------------------------------------------------------------------
void npCopyGPUtoCPU (void* dataRef)
{
	return;
}

// 2^24 max id for picking algorithm
//------------------------------------------------------------------------------
//MB-CAMERA
void npPick( int x, int y )
{
	int id = 0;

	GLubyte r,g,b,a;	//Unsigned Byte 8-bit Range: 0-255 // changed from u_int8_t, zz debug
	GLubyte pixels[4];		//4 Bytes
	
	NPnodePtr node = NULL;
	pData data = (pData) npGetDataRef();;
	
	
	glReadPixels( x,y, 1,1, GL_RGBA, GL_UNSIGNED_BYTE, pixels );			//Read the Pixel
	
	r = pixels[0];
	g = pixels[1];
	b = pixels[2];
	a = pixels[3];			// 2^24 rgb limit is not mapping alpha
	
	id = npRGBtoID( r,g,b );	//alpha currently not used, 2^24 rgb limit

	if (id >= kNPnodeMax)	//mouse selection crashing workaround, debug zz
		id = 0;				//add additional handling for bad picking IDs
		
	data->io.gl.pickID = id;

	if( id != 0 )
	{	
		node = data->map.nodeID[id];	//get node by id

		if (node != NULL)
		{
			if (data->io.mouse.buttonR)
				node->selected = 1 - node->selected;	//toggle value
			else
			{
				//if node not selected then clear selection, then select
				if (!node->selected)
					npCtrlCommand (kNPgraphSelectAllOff, data);

				node->selected = true;
			}

			//set active and selected node pointers
			npMapSelectNode (node, data);

			if (node->recordID)
				printf ("recordID: %d \n", node->recordID);
			else
				printf ("id: %d \n", node->id);
		}
	}
}

//MB-LABEL

//Draw "Ring" out of Toroid
//------------------------------------------------------------------------------
void drawRing()
{
//	glScalef( 12,12,12 );
	
	glScalef(100.0f,100.0f,100.0f);
	DrawSimpleTorus();
}


//Project CURRENT Matrix state World position Coordinates onto the Screen
//------------------------------------------------------------------------------
NPfloatXY npProjectWorldToScreen()
{	
	float pinOffset = 6.0f;

	GLint viewport[4];

	GLdouble screenX, screenY, dummyZ;	//Screen Coordinates Output

	NPfloatXY screenXY;

	GLdouble model_matrix[16];
	GLdouble projection_matrix[16];

	//Model Matrix Current
	glGetDoublev(GL_MODELVIEW_MATRIX, model_matrix);
	
	//Projection Matrix Current
	glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix);

	//Viewport
	glGetIntegerv(GL_VIEWPORT, viewport);	

	//Do the projection
	gluProject( 0,0, pinOffset, model_matrix, projection_matrix, viewport, &screenX, &screenY, &dummyZ );
	
//	WATCH_FLOAT( screenX );
//	WATCH_FLOAT( screenY );
//	WATCH_FLOAT( dummyZ  );

	screenXY.x = (float)screenX;
	screenXY.y = (float)screenY;

	return screenXY;
}

//Draw the Text Labels (Simple Ring now)
//------------------------------------------------------------------------------
void npDrawTextLabels()
{
	int i = 0;
	float sc;
	int nodeRootCount;
				
	NPfloatXY screenPosXY;

	pData data = (pData) npGetDataRef();
	NPnodePtr node;


//	WATCH_FUNC;
	
	glPushMatrix();
	
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );
	
	sc = 62.5f/data->io.gl.height;
	glScalef( sc, sc, sc );		//scale to Screen Coords
	
	nodeRootCount = data->map.nodeRootCount;
//	WATCH_INT( nodeRootCount );
	
	for( i = 0; i < nodeRootCount; i++ )
	{
		if( i < 3 )	//Nothing, Camera, Grid
			continue;
		
		node = data->map.node[i];					//change to sort[], debug zz
		
		screenPosXY = node->screenXY;
//		WATCH_VEC2( screenPosXY );
		screenPosXY.x -= data->io.gl.width  * 0.5f;
		screenPosXY.y -= data->io.gl.height * 0.5f;
		
		glPushMatrix();
			glTranslated( screenPosXY.x,screenPosXY.y,0 );	
			drawRing();
		glPopMatrix();
	}
	
	glPopMatrix();
}

